#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: psakic

This sub-module of geodezyx.reffram contains functions related to
 Kepler's element operations. 

it can be imported directly with:
from geodezyx import reffram

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""

########## BEGIN IMPORT ##########
#### External modules
import numpy as np
import pandas as pd
import datetime as dt
#from pytwobodyorbit import TwoBodyOrbit
from pytwobodyorbit import TwoBodyOrbit  

#### geodeZYX modules
from geodezyx import conv

#### Import the logger
import logging
log = logging.getLogger(__name__)

##########  END IMPORT  ##########


def ECI_2_kepler_elts(P,V,rad2deg=True,
                      mu=3.9860044188e14):
    """
    Convert ECI coordinates > Kepler's elements

    Parameters
    ----------
    P : array
        Position.
    V : array
        Velocity.
    rad2deg : bool, optional
        canvert Keplerian's angles (anomalies) to deg. The default is True.
    mu : float, optional
        Standard gravitational parameter. The default is 3.9860044188e14.

    Returns
    -------
    a,ecc,i,omega_periarg,omega_LAN,m : floats
        Kepler's elements.
    """
    Pnorm = np.linalg.norm(P)
    Vnorm = np.linalg.norm(V)
    
    ### orbital momentum vector H
    H = np.cross(P,V)
    Hnorm = np.linalg.norm(H)
    
    ### eccentricity vector Ecc
    Ecc = (np.cross(V,H)/mu) - (P/Pnorm)
    Ecc_norm = np.linalg.norm(Ecc)
    
    ### true anomaly ν (nu)
    N = np.cross(np.array([0,0,1]),H)
    Nnorm = np.linalg.norm(N)
    if np.dot(P,V) >= 0:
        nu = np.arccos(np.dot(Ecc,P)/(Ecc_norm*Pnorm))
    else:
        nu = 2*np.pi - np.arccos(np.dot(Ecc,P)/(Ecc_norm*Pnorm))        
    
    ### orbit inclination i
    i = np.arccos(H[2]/Hnorm)
    
    ### orbit eccentricity ecc
    ecc = np.linalg.norm(Ecc)
    
    ### eccentric anomaly E (called e here)
    #e1 = 2*np.arctan(np.tan(nu/2)/np.sqrt((1+ecc)/(1-ecc)))
    e2 = 2*np.arctan2(np.tan(nu/2),np.sqrt((1+ecc)/(1-ecc)))
    e=e2
    
    ### longitude of the ascending node Ω (omega_LAN)
    if N[1] >= 0:
        omega_LAN = np.arccos(N[0]/Nnorm)
    else:
        omega_LAN = 2*np.pi - np.arccos(N[0]/Nnorm)
    
    ### argument of periapsis ω (omega_periarg)
    if Ecc[2] >= 0:
        omega_periarg = np.arccos((np.dot(N,Ecc))/(Nnorm * Ecc_norm))
    else:
        omega_periarg = 2*np.pi - np.arccos((np.dot(N,Ecc))/(Nnorm * Ecc_norm))
    
    ### mean anomaly m, (Kepler’s Equation)
    m = e - ecc*np.sin(e)
    
    ### semi-major axis a
    a = ((2/Pnorm) - (Vnorm**2 / mu))**-1

    if rad2deg:
        i=np.rad2deg(i)
        omega_periarg=np.rad2deg(omega_periarg)
        omega_LAN=np.rad2deg(omega_LAN)
        m=np.rad2deg(m)

    return a,ecc,i,omega_periarg,omega_LAN,m



def extrapolate_orbit_kepler(P,V,t,t0,mu=3.9860044188e14):
    orbit = TwoBodyOrbit("", mu=mu)   # create an instance
    orbit.setOrbCart(t0, P, V)        # define the orbit
    Pout, Vout = orbit.posvelatt(t)   # get position and velocity at t1
    kepl = orbit.elmKepl()            # get classical orbital elements
    
    return Pout,Vout,kepl



def extrapolate_sp3_DataFrame(DFsp3,step=900,n_step=9,
                              backward=True,forward=True,
                              until_backward=None,
                              until_forward=None,
                              return_all=True):
    """
    Extrapolate the positions in a SP3 based on the first/last
    epochs' position

    Parameters
    ----------
    DFsp3 : DataFrame
        Input Orbit DataFrame (i.e. generated by files_rw.read_sp3).
    step : int, optional
        step between two epochs. The default is 900.
    n_step : int, optional
        number of epochs to interpolate. The default is 9.
    backward and forward : bool, optional
        extrapolate for the day before/after. The default is True.
    return_all : bool, optional
        if True, returns the input DF enhanced with the extrapolated values
        if False, returns only the extrapolated values.
        The default is True.
    until_backward & until_backward : datetime, optional
        epoch until then the extrapolation has to be done.
        Override n_step
        
    Returns
    -------
    DForb_out : DataFrame
        Orbit DataFrame with extrapolated values
        (see return_all).

    """
    NewEpoch_stk = []
    
    for sat in DFsp3["sat"].unique():
        log.info("extrapolate: %s",sat)
        DFsat = DFsp3[(DFsp3["sat"] == sat) & (DFsp3["type"] == "P")].copy()
        DFsat.sort_values("epoch",inplace=True)
        
        DFline_dummy = DFsat.iloc[0].copy()
        DFline_dummy["clk"] = 999999.999999
        
        XYZ   = DFsat[["x","y","z"]].values *1000
        Tgps = conv.numpy_dt2dt(DFsat["epoch"].values)
        Tutc = conv.dt_gpstime2dt_utc(Tgps)
        Tsec = [e.total_seconds() for e in (Tutc-Tutc[0])]
        
        P=conv.ECEF2ECI(XYZ,Tutc)
        V=np.gradient(P,Tsec,axis=0)
        
        def extrapo_intern_fct(i_ref,coef,until):
            ##
            ## backward :
            ## i_ref , coef = 0,-1
            ##
            ## forward :
            ## i_ref , coef = -1,1
            ##
            ## CORRDS ARE GIVEN IN ECI HERE !!!
            ##
            orbit_back = TwoBodyOrbit("orbit", mu=3.9860044188e14)  # create an instance    
            orbit_back.setOrbCart(Tsec[i_ref], P[i_ref], V[i_ref])  # define the orbit
            
            if until:
                t_rang_strt,t_rang_end = list(sorted([Tutc[i_ref],until]))
                Range = conv.dt_range(t_rang_strt,t_rang_end,0,step)
                n_step_intern = len(Range) - 1 
            else:
                n_step_intern = n_step
            
            for t in np.arange(step,n_step_intern*step +1,step):
                Pout, Vout = orbit_back.posvelatt(coef * t)
                epoc = Tgps[i_ref] + coef * dt.timedelta(seconds=int(t))
                DFline_new = DFline_dummy.copy()
                DFline_new[["x","y","z"]] = Pout
                DFline_new["epoch"] = pd.Timestamp(epoc)
        
                NewEpoch_stk.append(DFline_new)
                
            return Pout, Vout
    
        ### backward
        if backward:
            extrapo_intern_fct(0,-1,until_backward)
    
        ### forward
        if forward:
            extrapo_intern_fct(-1,1,until_forward)
        
    DFNewEpoch = pd.DataFrame(NewEpoch_stk)
    
    Tutc_NewEpoch = conv.dt_gpstime2dt_utc(conv.numpy_dt2dt(DFNewEpoch["epoch"].values))
    DFNewEpoch[["x","y","z"]] = conv.ECI2ECEF(DFNewEpoch[["x","y","z"]].values,Tutc_NewEpoch)
    DFNewEpoch[["x","y","z"]] = DFNewEpoch[["x","y","z"]] * 10**-3
    
    if return_all:
        DForb_out = pd.concat((DFsp3,DFNewEpoch))
    else:
        DForb_out = DFNewEpoch
    
    DForb_out.reset_index(drop=True,inplace=True)
    DForb_out.sort_values(["sat","epoch"],inplace=True)
    DForb_out.reset_index(drop=True,inplace=True)
    
    return DForb_out





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: psakic

This sub-module of geodezyx.reffram contains functions for low-level 
geometry operations. 

it can be imported directly with:
from geodezyx import reffram

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""

########## BEGIN IMPORT ##########
#### External modules
import io
import pandas as pd
import numpy as np
import scipy


#### geodeZYX modules
from geodezyx import conv
from geodezyx import stats
from geodezyx import utils

#### Import the logger
import logging
log = logging.getLogger(__name__)

##########  END IMPORT  ##########


#  _    _ _       _       _                    _    _____                _      _   _        ______   _       
# | |  | (_)     | |     | |                  | |  / ____|              | |    | | (_)      |  ____| | |      
# | |__| |_  __ _| |__   | |     _____   _____| | | |  __  ___  ___   __| | ___| |_ _  ___  | |__ ___| |_ ___ 
# |  __  | |/ _` | '_ \  | |    / _ \ \ / / _ \ | | | |_ |/ _ \/ _ \ / _` |/ _ \ __| |/ __| |  __/ __| __/ __|
# | |  | | | (_| | | | | | |___|  __/\ V /  __/ | | |__| |  __/ (_) | (_| |  __/ |_| | (__  | | | (__| |_\__ \
# |_|  |_|_|\__, |_| |_| |______\___| \_/ \___|_|  \_____|\___|\___/ \__,_|\___|\__|_|\___| |_|  \___|\__|___/
#            __/ |                                                                                            
#           |___/                                                                         
    
### High level geodetic function


def itrf_speed_calc(x0,y0,z0,t0,vx,vy,vz,t):
    """
    Translate a position to a given epoch using point velocity

    Parameters
    ----------
    x0,y0,z0 : float
        coordinates at the reference epoch (m).
    t0 : float
        reference epoch (decimal year).
    vx,vy,vz : float
        speed of the point (m/yr).
    t : float
        output epoch.

    Returns
    -------
    xout,yout,zout : floats
        coordinates of the point @ the ref. epoch (m).
    """
    

    xout = x0 + vx * ( t - t0 )
    yout = y0 + vy * ( t - t0 )
    zout = z0 + vz * ( t - t0 )

    return xout,yout,zout


def itrf_psd_fundamuntal_formula(t,A_l,t_l,tau_l,A_e,t_e,tau_e):
    """
    Get the Post Seismic Deformation of a station

    Parameters
    ----------
        
    t : float
        epoch in decimal years.
    A_l : float
        Amplitude of the logarithmic term.
    t_l : float
        Earthquake time(date) corresponding to logarithmic term
    tau_l : float
        Relaxation time of the logarithmic term.
    A_e : float
        Amplitude of the exponential term.
    t_e : float
        Earthquake time(date) corresponding to the exponential term.
    tau_e : float
        Relaxation time of the exponential term..
    
    Returns
    -------
    dL : TYPE
        the total sum of PSD corrections.

    References
    ----------
        http://itrf.ensg.ign.fr/ITRF_solutions/2014/doc/ITRF2014-PSD-model-eqs-IGN.pdf
    """
    
    dL = A_l * np.log(1 + (t - t_l)/ tau_l) + A_e * (1 + (t - t_e)/ tau_e)
    
    return dL


def calc_pos_speed_itrf(x0,y0,z0,t0,vx,vy,vz,t):
    """
    just a wrapper of itrf_speed_calc
    for legacy reasons
    """
    return itrf_speed_calc(x0,y0,z0,t0,vx,vy,vz,t)


def itrf_helmert_get_parameters(TRF_Input_name,TRF_Ext_name,
                                verbose=False,
                                convert=True):
    """
    Get the Helmert parameters for a transformation I/ETRFxx <=> I/ETRFxx

    Parameters
    ----------
    TRF_Input_name : str
        Name of the initial Reference Frame.
        Can be ITRFyyyy, ITRFyy, ETRFyyyy, ETRFyy
        where yyyy or yy is the TRF realization
    TRF_Ext_name : str
        Name of the wished Reference Frame.
        Can be ITRFyyyy, ITRFyy, ETRFyyyy, ETRFyy
        where yyyy or yy is the TRF realization
    verbose : bool, optional
        Print the parameters. The default is True.
    convert : bool, optional
        Gives directly useful values in the good units
        for a proper transformation

    Returns
    -------
    T : 3-Array
        Translation.
    Tdot : 3-Array
        Translation rate.
    D : float
        Scale factor.
    Ddot : float
        Scale factor rate.
    R : 3-Array
        Rotation.
    Rdot : 3-Array
        Rotation rte.
    epoch_ref_Xe : float
        Reference epoch of TRF_Ext_name.
        
    Notes
    -----
    Based on the values of
    http://etrs89.ensg.ign.fr/pub/EUREF-TN-1.pdf
    
    Warning: a validation unsing the EUREF converter is highly recomended
    https://www.epncb.oma.be/_productsservices/coord_trans/
    

    """

    ### Get short name if necessary
    if len(TRF_Input_name) == 8:
        TRF_Input_name = TRF_Input_name[:4] + TRF_Input_name[6:]
    if len(TRF_Ext_name) == 8:
        TRF_Ext_name = TRF_Ext_name[:4] + TRF_Ext_name[6:]
        
        
    COEFFS="""TRF_I,TRF_E,TRF_E_epoch,T1,T2,T3,D,R1,R2,R3,Tdot1,Tdot2,Tdot3,Ddot,Rdot1,Rdot2,Rdot3
ITRF14,ETRF14,1989.0,0.0,0.0,0.0,0.00,0.000,0.000,0.000,0.0,0.0,0.0,0.00,0.085,0.531,-0.770
ITRF05,ETRF05,1989.0,56.0,48.0,-37.0,0.00,0.000,0.000,0.000,0.0,0.0,0.0,0.00,0.054,0.518,-0.781
ITRF00,ETRF00,1989.0,54.0,51.0,-48.0,0.00,0.000,0.000,0.000,0.0,0.0,0.0,0.00,0.081,0.490,-0.792
ITRF97,ETRF97,1989.0,41.0,41.0,-49.0,0.00,0.000,0.000,0.000,0.0,0.0,0.0,0.00,0.200,0.500,-0.650
ITRF96,ETRF96,1989.0,41.0,41.0,-49.0,0.00,0.000,0.000,0.000,0.0,0.0,0.0,0.00,0.200,0.500,-0.650
ITRF94,ETRF94,1989.0,41.0,41.0,-49.0,0.00,0.000,0.000,0.000,0.0,0.0,0.0,0.00,0.200,0.500,-0.650
ITRF93,ETRF93,1989.0,19.0,53.0,-21.0,0.00,0.000,0.000,0.000,0.0,0.0,0.0,0.00,0.320,0.780,-0.670
ITRF92,ETRF92,1989.0,38.0,40.0,-37.0,0.00,0.000,0.000,0.000,0.0,0.0,0.0,0.00,0.210,0.520,-0.680
ITRF91,ETRF91,1989.0,21.0,25.0,-37.0,0.00,0.000,0.000,0.000,0.0,0.0,0.0,0.00,0.210,0.520,-0.680
ITRF90,ETRF90,1989.0,19.0,28.0,-23.0,0.00,0.000,0.000,0.000,0.0,0.0,0.0,0.00,0.110,0.570,-0.710
ITRF89,ETRF89,1989.0,0.0,0.0,0.0,0.00,0.000,0.000,0.000,0.0,0.0,0.0,0.00,0.110,0.570,-0.710
ITRF14,ETRF14,2010.0,0.0,0.0,0.0,0.00,1.785,11.151,-16.170,0.0,0.0,0.0,0.00,0.085,0.531,-0.770
ITRF08,ETRF14,2010.0,-1.6,-1.9,-2.4,0.02,1.785,11.151,-16.170,0.0,0.0,0.1,-0.03,0.085,0.531,-0.770
ITRF05,ETRF14,2010.0,-2.6,-1.0,2.3,-0.92,1.785,11.151,-16.170,-0.3,0.0,0.1,-0.03,0.085,0.531,-0.770
ITRF00,ETRF14,2010.0,-0.7,-1.2,26.1,-2.12,1.785,11.151,-16.170,-0.1,-0.1,1.9,-0.11,0.085,0.531,-0.770
ITRF97,ETRF14,2010.0,-7.4,0.5,62.8,-3.80,1.785,11.151,-16.430,-0.1,0.5,3.3,-0.12,0.085,0.531,-0.790
ITRF96,ETRF14,2010.0,-7.4,0.5,62.8,-3.80,1.785,11.151,-16.430,-0.1,0.5,3.3,-0.12,0.085,0.531,-0.790
ITRF94,ETRF14,2010.0,-7.4,0.5,62.8,-3.80,1.785,11.151,-16.430,-0.1,0.5,3.3,-0.12,0.085,0.531,-0.790
ITRF93,ETRF14,2010.0,50.4,-3.3,60.2,-4.29,4.595,14.531,-16.570,2.8,0.1,2.5,-0.12,0.195,0.721,-0.840
ITRF92,ETRF14,2010.0,-15.4,-1.5,70.8,-3.09,1.785,11.151,-16.430,-0.1,0.5,3.3,-0.12,0.085,0.531,-0.790
ITRF91,ETRF14,2010.0,-27.4,-15.5,76.8,-4.49,1.785,11.151,-16.430,-0.1,0.5,3.3,-0.12,0.085,0.531,-0.790
ITRF90,ETRF14,2010.0,-25.4,-11.5,92.8,-4.79,1.785,11.151,-16.430,-0.1,0.5,3.3,-0.12,0.085,0.531,-0.790
ITRF89,ETRF14,2010.0,-30.4,-35.5,130.8,-8.19,1.785,11.151,-16.430,-0.1,0.5,3.3,-0.12,0.085,0.531,-0.790
ITRF14,ETRF00,2010.0,54.7,52.2,-74.1,2.12,1.701,10.290,-16.632,0.1,0.1,-1.9,0.11,0.081,0.490,-0.792
ITRF08,ETRF00,2010.0,53.1,50.3,-76.5,2.14,1.701,10.290,-16.632,0.1,0.1,-1.8,0.08,0.081,0.490,-0.792
ITRF05,ETRF00,2010.0,52.1,51.2,-71.8,1.20,1.701,10.290,-16.632,-0.2,0.1,-1.8,0.08,0.081,0.490,-0.792
ITRF00,ETRF00,2010.0,54.0,51.0,-48.0,0.00,1.701,10.290,-16.632,0.0,0.0,0.0,0.00,0.081,0.490,-0.792
ITRF97,ETRF00,2010.0,47.3,52.7,-11.3,-1.68,1.701,10.290,-16.892,0.0,0.6,1.4,-0.01,0.081,0.490,-0.812
ITRF96,ETRF00,2010.0,47.3,52.7,-11.3,-1.68,1.701,10.290,-16.892,0.0,0.6,1.4,-0.01,0.081,0.490,-0.812
ITRF94,ETRF00,2010.0,47.3,52.7,-11.3,-1.68,1.701,10.290,-16.892,0.0,0.6,1.4,-0.01,0.081,0.490,-0.812
ITRF93,ETRF00,2010.0,105.1,48.9,-13.9,-2.17,4.511,13.670,-17.032,2.9,0.2,0.6,-0.01,0.191,0.680,-0.862
ITRF92,ETRF00,2010.0,39.3,50.7,-3.3,-0.97,1.701,10.290,-16.892,0.0,0.6,1.4,-0.01,0.081,0.490,-0.812
ITRF91,ETRF00,2010.0,27.3,36.7,2.7,-2.37,1.701,10.290,-16.892,0.0,0.6,1.4,-0.01,0.081,0.490,-0.812
ITRF90,ETRF00,2010.0,29.3,40.7,18.7,-2.67,1.701,10.290,-16.892,0.0,0.6,1.4,-0.01,0.081,0.490,-0.812
ITRF89,ETRF00,2010.0,24.3,16.7,56.7,-6.07,1.701,10.290,-16.892,0.0,0.6,1.4,-0.01,0.081,0.490,-0.812
ITRF14,ITRF08,2010.0,1.6,1.9,2.4,-0.02,0.00,0.00,0.00,0.0,0.0,-0.1,0.03,0.00,0.00,0.00
ITRF14,ITRF05,2010.0,2.6,1.0,-2.3,0.92,0.00,0.00,0.00,0.3,0.0,-0.1,0.03,0.00,0.00,0.00
ITRF14,ITRF00,2010.0,0.7,1.2,-26.1,2.12,0.00,0.00,0.00,0.1,0.1,-1.9,0.11,0.00,0.00,0.00
ITRF14,ITRF97,2010.0,7.4,-0.5,-62.8,3.80,0.00,0.00,0.26,0.1,-0.5,-3.3,0.12,0.00,0.00,0.02
ITRF14,ITRF96,2010.0,7.4,-0.5,-62.8,3.80,0.00,0.00,0.26,0.1,-0.5,-3.3,0.12,0.00,0.00,0.02
ITRF14,ITRF94,2010.0,7.4,-0.5,-62.8,3.80,0.00,0.00,0.26,0.1,-0.5,-3.3,0.12,0.00,0.00,0.02
ITRF14,ITRF93,2010.0,-50.4,3.3,-60.2,4.29,-2.81,-3.38,0.40,-2.8,-0.1,-2.5,0.12,-0.11,-0.19,0.07
ITRF14,ITRF92,2010.0,15.4,1.5,-70.8,3.09,0.00,0.00,0.26,0.1,-0.5,-3.3,0.12,0.00,0.00,0.02
ITRF14,ITRF91,2010.0,27.4,15.5,-76.8,4.49,0.00,0.00,0.26,0.1,-0.5,-3.3,0.12,0.00,0.00,0.02
ITRF14,ITRF90,2010.0,25.4,11.5,-92.8,4.79,0.00,0.00,0.26,0.1,-0.5,-3.3,0.12,0.00,0.00,0.02
ITRF14,ITRF89,2010.0,30.4,35.5,-130.8,8.19,0.00,0.00,0.26,0.1,-0.5,-3.3,0.12,0.00,0.00,0.02
ITRF14,ITRF88,2010.0,25.4,-0.5,-154.8,11.29,0.10,0.00,0.26,0.1,-0.5,-3.3,0.12,0.00,0.00,0.02
"""
    
    COEFFS = io.StringIO(COEFFS)
    #p="/home/psakicki/Downloads/coeff_helmert2.csv"
    DF = pd.read_csv(COEFFS)
        
    DF2 = DF[(DF["TRF_I"] == TRF_Input_name) & (DF["TRF_E"] == TRF_Ext_name)]
    inver = 1.
    inver_bool = False
    
    if len(DF2) == 0:
        DF2 = DF[(DF["TRF_E"] == TRF_Input_name) & (DF["TRF_I"] == TRF_Ext_name)]
        inver = -1.
        inver_bool = True
    
    DF3 = DF2.iloc[0]
    
    if convert:
        convT = 0.001
        convD = 10**-9
        convR = 4.84813681109536e-09
    else:
        convT = 0.
        convD = 0.
        convR = 0.
    
    T    = inver * convT * DF3[["T1","T2","T3"]].values
    Tdot = inver * convT * DF3[["Tdot1","Tdot2","Tdot3"]].values
    
    D    = inver * convD * DF3["D"]
    Ddot = inver * convD * DF3["Ddot"]
    
    R    = inver * convR * DF3[["R1","R2","R3"]].values
    Rdot = inver * convR * DF3[["Rdot1","Rdot2","Rdot3"]].values
    
    epoch_ref_Xe = DF3["TRF_E_epoch"]
    
    if verbose:
        log.info("%s => %s",TRF_Input_name,TRF_Ext_name)
        log.info("   T: %s",T,Tdot)
        log.info("   D: %s",D,Ddot)
        log.info("   R: %s",R,Rdot)
        log.info("t Xe: %s",epoch_ref_Xe)
        log.info(" Inv: %s",inver_bool)
        log.info("DataFrame:")
        log.info(DF2)
    
    return T,Tdot,D,Ddot,R,Rdot,epoch_ref_Xe


def itrf_helmert_trans(Xi,
                       epoch_Xi,
                       T,Tdot,
                       D,Ddot,
                       R,Rdot,
                       epoch_ref_Xe):
    """
    Do the Helmert transformation I/ETRFxx <=> I/ETRFxx
    
    The seven Helmert Parameters are generated with ``itrf_helmert_get_parameters``

    Parameters
    ----------
    Xi : 3-Array, Nx3-Array OR 6-Array, Nx6-Array
        Points in the initial Reference Frame.
        3 columns if positions only
        6 columns if positions+velocities
    epoch_Xi : float
        epoch of the initial Reference Frame points (decimal year).
    T : 3-Array
        Translation parameter.
    Tdot : 3-Array
        Translation rate parameter.
    D : float
        Scale factor parameter.
    Ddot : float
        Scale factor rate parameter.
    R : 3-Array
        Rotation parameter.
    Rdot : 3-Array
        Rotation rate parameter.
    epoch_ref_Xe : float
        Reference epoch of the destination TRF.

    Returns
    -------
    Xe : 3-Array or Nx3-Array
        Points in the destination Reference Frame.
        

        
    Notes
    -----
    
    ** Warning **
    This function does not the velocity shift from one epoch to another \n
    The output coordinates will be provided at the same epoch as the input ones \n
    Use ``itrf_speed_calc`` function to perform this potential step 
    before (with the initial velocities) or after (with the transformed velocities)
    calling the present function.
    
    
    Based on the theory and values of *EUREF Technical Note 1: 
    Relationship and Transformation between the International 
    and the European Terrestrial Reference Systems, Z. Altamimi, 2018*
    http://etrs89.ensg.ign.fr/pub/EUREF-TN-1.pdf
    
    We recommend to confirm the values with the official EUREF converter
    http://www.epncb.oma.be/_productsservices/coord_trans/index.php
    
    
    ** Notes for the French geodetic system **    
    By definition, since the 2021/01/04
    RGF93(v2b) = ETRF2000@2019.0
    
    References
    ----------
    https://geodesie.ign.fr/index.php?page=rgf93 \n
    https://geodesie.ign.fr/contenu/fichiers/RGF93v2b-RAF18b.pdf \n
    https://geodesie.ign.fr/contenu/fichiers/rgf93v2b_information_cnig.pdf 
    
    """
    ## prelimiary warning 
    if utils.is_iterable(epoch_Xi):
        log.warning("epoch_Xi is an iterable !!!")
        log.info("The function works correctly with a single epoch only !!!")
        log.info("if several initial Reference Frame epoch")
        log.info("use a loop outside the function")
        
    
    if not type(Xi) is np.array:
        Xi = np.array(Xi)
    
    # manage the case where only one point is given
    if len(np.shape(Xi)) == 1:
        Xi = Xi[...,np.newaxis].T
        
    
    ##### Here we manage if there is velocities or not 
    if Xi.shape[1] == 6:
        velocity_mode = True     
    else:
        velocity_mode = False
        
        
    #### Trick if we are in velocity mode, we recall the fct 
    #### with velocity_mode=False to have the positions
    if velocity_mode:
        Xe_pos = itrf_helmert_trans(Xi[:,:3],epoch_Xi,
                                     T,Tdot,
                                     D,Ddot,
                                     R,Rdot,
                                     epoch_ref_Xe)
        

    #### Translation
    if velocity_mode:
        Topera = Tdot
    else:
        Topera = T + Tdot * (epoch_Xi-epoch_ref_Xe)
        
    #### Scale Factor    
    D_mat       = np.eye(3).dot(D)
    Ddot_mat    = np.eye(3).dot(Ddot)
    if velocity_mode:
        D_mat_opera = Ddot_mat
    else:
        D_mat_opera = Ddot_mat * (epoch_Xi-epoch_ref_Xe) + D_mat
        
    ### Rotation
    r_1,r_2,r_3 = R  
    r_dot_1,r_dot_2,r_dot_3 = Rdot 
    
    R_mat    = np.array([[0. ,-r_3, r_2],
                         [ r_3, 0.,-r_1],
                         [-r_2, r_1,0.]])     
    
    Rdot_mat = np.array([[0. ,-r_dot_3, r_dot_2],
                         [ r_dot_3, 0.,-r_dot_1],
                         [-r_dot_2, r_dot_1,0.]])     
    
    if velocity_mode:
        R_mat_opera = (Rdot_mat)     
    else:
        R_mat_opera = (Rdot_mat * (epoch_Xi-epoch_ref_Xe) + R_mat) 


    #### Computation
    Xe_stk = []
    for xi in Xi:
        if velocity_mode:
            xe = xi[3:] + Topera + D_mat_opera.dot(xi[:3]) + R_mat_opera.dot(xi[:3])
        else:
            xe = xi[:3] + Topera + D_mat_opera.dot(xi[:3]) + R_mat_opera.dot(xi[:3])            

        Xe_stk.append(xe)

    #### Final Aestetics
    Xe = np.array(Xe_stk)
    Xe = np.squeeze(Xe)
    Xe = Xe.astype(np.float64)
    
    ### final concat of the position and the velocities
    if velocity_mode:
         if len(np.shape(Xe)) == 1:
             Xe = np.hstack((Xe_pos,Xe))
         else:
             Xe = np.column_stack((Xe_pos,Xe))

    return Xe



def _helmert_trans_estim_matrixs_maker(X1 , X2):
    """
    internal function for helmert_trans_estim
    """
    x1 , y1 , z1 = X1
    x2 , y2 , z2 = X2
    
    
    block_1 = np.eye(3)
    
    block_2 = np.array([[ 0. , -z1,  y1, x1],
                        [ z1,   0., -x1, y1],
                        [-y1,  x1,   0., z1]])
    
    l = X2 - X1
    A = np.hstack((block_1 , block_2))

    return l,A


# def helmert_trans_frontend(X1list , X2list, Weights=[]):
#     HParam , A , l = reffram.helmert_trans_estim(X1list,X2list)
#     X2_trans_out = reffram.helmert_trans_apply(X1list,HParam)
  

def helmert_trans_estim(X1list , X2list, Weights=[]):
    """
    estimates 7 parameters of a 3D Helmert transformation between a set of points
    X1 and a set of points X2 (compute transformation X1 => X2)
    
    Parameters
    ----------
    
    X1list & X2list : list of N (x,y,z) points, or an (N,3)-shaped numpy array
        Input point sets

    Weights : list of N Weights,
        or an numpy array of shape (N,3)
    
    Returns
    -------
    
    HParam :
        7 Helmert params. : x,y,z translations, x,y,z rotations, scale
    A :
        Design matrix    
    l :
        Differences X2 - X1 (before transformation !!!)
        
    References
    ----------
    https://elib.uni-stuttgart.de/bitstream/11682/9661/1/BscThesis_GaoYueqing.pdf
    """

    l_stk = []
    A_stk = []
    Bool_stk = []
    
    log.error("we have %s points",len(X1list))
    
    for X1 , X2 in zip(X1list , X2list):
        
        if np.sum(np.isnan(X1)) or np.sum(np.isnan(X2)):
            log.warning("one point component is nan, skipping")
            Bool_stk.append(False)
            continue
        
        lmono , Amono = _helmert_trans_estim_matrixs_maker(X1,X2)
            
        l_stk.append(lmono)
        A_stk.append(Amono)
        Bool_stk.append(True)
        
        
    A = np.vstack(A_stk)
    l = np.hstack(l_stk)
    Bool = np.array(Bool_stk)
    
    if len(Weights) == 0:
        W = np.eye(len(l))
    else:
        Weights_corr = np.array(Weights)
        Weights_corr = Weights_corr[Bool]
        W = np.repeat(Weights_corr,3) * np.eye(len(l))
    
    N    = (A.T).dot(W).dot(A)
    AtWB = (A.T).dot(W).dot(l)
    
    HParam = scipy.linalg.inv(N).dot(AtWB)
    
    return HParam , A , l


def helmert_trans_apply(Xin,SevenParam_in,legacy_mode=False):
    """
    Apply an Helmert transformation (7-parameters)
    to a set of points

    Parameters
    ----------
    Xin : list of N (x,y,z) points, or an (N,3)-shaped numpy array.
        input set points
        
    SevenParam_in : 7 element list or array
        7 Helmert params. : x,y,z translations, x,y,z rotations, scale.
        
    legacy_mode : bool, optional
        Use a non-optimized and slow computation approach (but same result).
        This option should be removed in the Future.
        The default is False.

    Returns
    -------
    Xout : list/array of N (x,y,z) points
        output transformed points. Same type as the input.

    """
    
    tx,ty,tz,rx,ry,rz,scal = SevenParam_in 
    
    R = np.array([[1.,rz,-ry],
                  [-rz,1.,rx],
                  [ry,-rx,1.]])
    
    T = np.array([tx,ty,tz])
    S = (1. + scal)
    
    typ=utils.get_type_smart(Xin)
    
    #### Apply the transformation here 
    if legacy_mode: #### SLOW !!!
        Xout = []
        for X1 in Xin:
            X2 = S * np.dot(R,X1) + T
            Xout.append(X2)
    else: ##### 100x faster with the Einstein sum
        Xout = S * np.einsum('ij,kj->ki', R,Xin) + np.tile(T,(len(Xin),1))
  
    Xout=typ(Xout)
    
    return Xout



def helmert_trans_estim_minimisation(X1in,X2in,HParam_apri=np.zeros(7),
                                     L1norm=True,tol=10**-9,
                                     full_output=False,method="Powell"):    
    """
    estimates 7 parameters of a 3D Helmert transformation between a set of points
    X1 and a set of points X2 (compute transformation X1 => X2) 
    using a Minimization approach (and not a Least Square inversion)
    
    Parameters
    ----------
    
    X1in & X2in : list of N (x,y,z) points, or an (N,3)-shaped numpy array
        Input point sets

    HParam_apri : list of 7 values,
        The Apriori for the Helmert parameter 
    
    L1norm : bool
        Use the L1-norm as a criteria, use the quadratic sum instead if False
        
    tol : float
        tolerence for the convergence
    
    full_output : bool
        return only the result if True, return the scipy optimize result if False
    
    method : str, optional
        minimization method.
        see scipy.optimize.minimize for details
        The default is "Powell".
    
    Returns
    -------
    Res :
        7 Helmert params. : x,y,z translations, x,y,z rotations, scale
    """
    
    def minimiz_helmert_fct(HParam_mini_in,X1in,X2in,L1norm_mini=L1norm):
        """
        This fct is the input for the scipy optimization fct
        """
        HParam_mini_wk = HParam_mini_in.copy()
        X12wrk = helmert_trans_apply(X1in,HParam_mini_wk)

        if not L1norm_mini: #return a L2 norm 
            SUM=np.sum((np.sum(np.power(X2in - X12wrk,2),axis=1)))
        else: #Return L1 norm
            SUM=np.sum(np.sum(np.abs(X2in - X12wrk),axis=1))
        return SUM
        
    RES = scipy.optimize.minimize(minimiz_helmert_fct,HParam_apri,
                                  (X1in,X2in,L1norm),
                                  method=method,tol=tol,
                                  options={"maxiter":1000,
                                           'xtol':tol,
                                           'ftol':tol})
    
    if RES.status != 0:
        log.warning("something went wrong (status != 0)")
        log.warning("here is the scipy.optimize.minimize message")
        log.warning(" > " + RES.message)
    
    if not full_output:
        return RES.x
    else:
        return RES
    
def helmert_trans_estim_minimisation_scalar(X1,X2,HParam_opti_apriori,
                                            L1norm=True,itera=2):
    """
    Estimates the Helmert parameters but based on a minimisation approach 
    between X1 and X2 (as suggested in the IGS combination software)
    
    NOT STABLE AVOID THE USE
    (it does the optimization of the 1st parameter, 
    and then the optimization of the 2nd one etc, etc...)

    Parameters
    ----------
    X1 & X2 : list of N (x,y,z) points, or an (N,3)-shaped numpy array
        Input point sets
    HParam_opti_apriori : list of 7 values
        The Apriori for the Helmert parameter 
    L1norm : bool
        Use the L1-norm as a criteria, use the quadratic sum instead if False
    itera : int, optional
        number of iterations. The default is 2.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """

    
    log.warning("unstable approach, avoid !!!!!")
    
    def minimiz_helmert_fct_scalar(hparam_mono_in,hparam_mono_id,
                            HParam_mini_in,X1in,X2in):
        """
        This fct is the input for the scipy optimization fct
        """
        HParam_mini_wk = HParam_mini_in.copy()
        HParam_mini_wk[hparam_mono_id] = hparam_mono_in
        X12wrk = helmert_trans_apply(X1in,HParam_mini_wk)

        if not L1norm: #return a L2 norm 
            return np.sum(np.sqrt(np.sum(np.power(X2in - X12wrk,2),axis=1)))
        else: #Return L1 norm
            return np.sum(np.sum(np.abs(X2in - X12wrk),axis=1))
        
    HParam_opti_wrk = HParam_opti_apriori.copy()

    for j in range(itera): #iter iterations
        for i in range(7): #7 Helmert Parameters:
            RES=scipy.optimize.minimize_scalar(minimiz_helmert_fct_scalar,
                                               args=(i,HParam_opti_wrk,X1,X2),
                                               tol=10**-20)
            HParam_opti_wrk[i] = RES.x
            
        if not j:
            HParam_opti_prev = HParam_opti_wrk.copy()
        else:
            log.info("Helmert Param minimisation iter %s %s",j+1,HParam_opti_wrk - HParam_opti_prev)
            HParam_opti_prev = HParam_opti_wrk.copy()
                        
    return HParam_opti_wrk

        
#### ASTRONOMY FUNCTION
def semi_major_axis_from_mean_motion(n):
    """
    source : https://space.stackexchange.com/questions/18289/how-to-get-semi-major-axis-from-tle
    """
    mu = 3.9860044189 * 10**14
    a  = (mu**(1./3.)) / ((2*n*np.pi/86400)**(2./3.))
    return a    

 #  _                     _                    _    _____                           _        _        ______                _   _                 
 # | |                   | |                  | |  / ____|                         | |      (_)      |  ____|              | | (_)                
 # | |     _____      __ | |     _____   _____| | | |  __  ___  ___  _ __ ___   ___| |_ _ __ _  ___  | |__ _   _ _ __   ___| |_ _  ___  _ __  ___ 
 # | |    / _ \ \ /\ / / | |    / _ \ \ / / _ \ | | | |_ |/ _ \/ _ \| '_ ` _ \ / _ \ __| '__| |/ __| |  __| | | | '_ \ / __| __| |/ _ \| '_ \/ __|
 # | |___| (_) \ V  V /  | |___|  __/\ V /  __/ | | |__| |  __/ (_) | | | | | |  __/ |_| |  | | (__  | |  | |_| | | | | (__| |_| | (_) | | | \__ \
 # |______\___/ \_/\_/   |______\___| \_/ \___|_|  \_____|\___|\___/|_| |_| |_|\___|\__|_|  |_|\___| |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
                                                                                                                                                
                                                                                                                                                
def BL_from_points(listpointin):
    """  
    From a list of 2-D or 3-dD points, returns the a matrix with distance 
    between each points 
    
    Parameters
    ----------
    listpointin : list or numpy.array
        List of N 2D or 3D points [[x1,y1,z1] ... [xn , yn , zn]]
                
    Returns
    -------
    BL : numpy.array
        matrix with distances between each points
        
    """

    N = len(listpointin)
    BL = np.empty((N,N))

    for i,pt1 in enumerate(listpointin):
        for j,pt2 in enumerate(listpointin):

            if i == j:
                BL[i,j] = 0
            else:
                BL[i,j] = np.linalg.norm(pt1 - pt2)

    return BL


def rotmat2(theta,angtype='deg'):

    if angtype == 'deg':
        theta = np.deg2rad(theta)

    rotmat = np.array([[np.cos(theta),-np.sin(theta)],
                       [np.sin(theta),np.cos(theta)]])

    return rotmat


def rotmat3(alpha,beta,gamma,xyzreftuple = ([1, 0, 0], [0, 1, 0], [0, 0, 1]),angtype='deg'):

    xaxis, yaxis, zaxis = xyzreftuple

    if angtype == 'deg':
        alpha = np.deg2rad(alpha)
        beta  = np.deg2rad(beta)
        gamma = np.deg2rad(gamma)

    Rx = trans.rotation_matrix(alpha, xaxis)
    Ry = trans.rotation_matrix(beta, yaxis)
    Rz = trans.rotation_matrix(gamma, zaxis)
    R  = trans.concatenate_matrices(Rz, Ry, Rx)[:3,:3]

    return R

def rotate_points(alphal,betal,gammal,pointlin,Rtype='R1',
                  xyzreftuple = ([1, 0, 0], [0, 1, 0], [0, 0, 1]),
                  angtype='deg',fullout = False):
    '''
    R1  = Rz(g) * Ry(b) * Rx(a)
         si les RPY sont donnés dans le NED
         alors les positions résultantes sont dans le NED

    R2  =  matrice RPY2ENU
        si les RPY sont donnés dans le NED
        alors les  résultantes sont DANS LE ENU
        pas besoin de rotation NED2ENU

        Grewal et al. 2007

    Entrée :
        Angles n = A
        liste de listes de P * [ points ]

    Sortie :
        liste de listes [ [ xA ] [ xA ] ... xP [ xA ] ] 
    '''

    xaxis, yaxis, zaxis = xyzreftuple

    if not utils.is_iterable(alphal):
        alphal = np.array([alphal])
        betal = np.array([betal])
        gammal = np.array([gammal])
        boolnotiterable = True
    else:
        boolnotiterable = False

    pointlout = []
    R_out = []


    for pt in pointlin:

        if not utils.is_iterable(pt) or len(pt) != 3:
            log.error("pts != 3 coords")
            return 0

        pointltmp = []

        for a,b,g in zip(alphal,betal,gammal):

            R1 = rotmat3(a,b,g,angtype=angtype,xyzreftuple=xyzreftuple)
            R2 = conv.C_rpy2enu(a,b,g,angtype=angtype)

            if Rtype == 'R1':
                R = R1
            elif Rtype == 'R2':
                R = R2
            R_out.append(R)

            pointltmp.append(np.dot(R,pt))

        pointlout.append(pointltmp)

        if boolnotiterable:
            pointlout = pointltmp

        pointlout = np.array(pointlout)

    if fullout:
        return pointlout , R_out
    else:
        return pointlout


def guess_seq_len(seq):
    #source
    #http://stackoverflow.com/questions/11385718/python-finding-repeating-sequence-in-list-of-integers
    guess = 1

    if len(set(seq)) == 1:
        return 1

    max_len = len(seq) / 2
    for x in range(2, max_len):
        if seq[0:x] == seq[x:2*x] :
            return x

    return guess

def wrapTo2Pi(lon):
    """
     wrapTo2Pi Wrap angle in radians to [0 2*pi]

    lambdaWrapped = wrapTo2Pi(LAMBDA) wraps angles in LAMBDA, in radians,
    to the interval [0 2*pi] such that zero maps to zero and 2*pi maps
    to 2*pi. (In general, positive multiples of 2*pi map to 2*pi and
    negative multiples of 2*pi map to zero.)

    See also wrapToPi, wrapTo180, wrapTo360.

    """
    lon = np.array(lon)
    positiv = lon > 0
    outlon = np.mod(lon , 2*np.pi)
    outlon[np.logical_and(outlon == 0 , positiv)] = 2 * np.pi
    return outlon


def wrapToPi(lon):
    """
    wrapToPi Wrap angle in radians to [-pi pi]

       lambdaWrapped = wrapToPi(LAMBDA) wraps angles in LAMBDA, in radians,
       to the interval [-pi pi] such that pi maps to pi and -pi maps to
       -pi.  (In general, odd, positive multiples of pi map to pi and odd,
       negative multiples of pi map to -pi.)

       See also wrapTo2Pi, wrapTo180, wrapTo360.

    """

    outlon = np.array(lon)
    q =  np.logical_and((outlon < -np.pi) , (np.pi < outlon))
    outlon[q] = wrapTo2Pi(outlon[q] + np.pi) - np.pi
    return outlon


def wrapTo180(lonin):
    """
    wrapTo180 Wrap angle in degrees to [-180 180]

    lonWrapped = wrapTo180(LON) wraps angles in LON, in degrees, to the
    interval [-180 180] such that 180 maps to 180 and -180 maps to -180.
    (In general, odd, positive multiples of 180 map to 180 and odd,
    negative multiples of 180 map to -180.)

    See also wrapTo360, wrapTo2Pi, wrapToPi.
    """
    lon = np.array(lonin)
    q = (lon < -180) and (180 < lon)
    lon[q] = wrapTo360(lon[q] + 180) - 180

    return lon

def wrapTo360(lonin):
    """
    wrapTo360 Wrap angle in degrees to [0 360]

    lonWrapped = wrapTo360(LON) wraps angles in LON, in degrees, to the
    interval [0 360] such that zero maps to zero and 360 maps to 360.
    (In general, positive multiples of 360 map to 360 and negative
    multiples of 360 map to zero.)

    See also wrapTo180, wrapToPi, wrapTo2Pi.
    """

    lon = np.array(lonin)

    positiveInput = (lon > 0)
    lon = np.mod(lon, 360)
    lon[(lon == 0) & positiveInput] = 360
    return lon



# Pas convaincu de son utilité
def unwrap180(anglist,angtype='deg'):

    if angtype == 'deg':
        seuil = 360

    angout = []

    for a in anglist:
        if a > seuil / 2:
            a = a - seuil
        angout.append(a)

    return angout


def wrap360(anglist,angtype='deg'):

    angout = []

    if angtype == 'deg':
        seuil = 360
    elif angtype == 'rad':
        seuil = 2*np.pi

    for a in anglist:
        if a < 0:
            a = a + seuil

        angout.append(a)

    return angout

class interp1d_ang():

    def __init__(self,T,A,angtype='deg',kind='linear',bounds_error=False):

        if angtype == 'deg':
            A = np.deg2rad(A)

        self.A = A
        self.T = T
        self.C = np.cos(A)
        self.S = np.sin(A)

        self.CfT = scipy.interpolate.interp1d(T,self.C,kind=kind,bounds_error=bounds_error)
        self.SfT = scipy.interpolate.interp1d(T,self.S,kind=kind,bounds_error=bounds_error)


    def __call__(self,T,angtype='deg'):

        I = np.arctan2(self.SfT(T) ,self.CfT(T) )
        I = wrap360(I,angtype='rad')

        if angtype == 'deg':
            return np.rad2deg(I)
        else:
            return I

def group_consecutives(vals, step=1):
    """
    Return list of consecutive lists of numbers from vals (number list).
    """
    run = []
    result = [run]
    expect = None
    for v in vals:
        if (v == expect) or (expect is None):
            run.append(v)
        else:
            run = [v]
            result.append(run)
        expect = v + step

        result2 = []
        for r in result:
            if len(r) > 1:
                result2.append([r[0],r[-1]])
            else:
                result2.append(r)

    return result2

def randomwalk_normal(N=100, d=2 , moy = 0 , sigma = 1):
    """
    d = dimension
    """
    return np.cumsum(moy + np.random.randn(N,d) * sigma)

def randomwalk_uniform(N=100, d=2 , bound = 0.5):
    """
    d = dimension
    bound = contraint of the random walk
    """
    return np.cumsum(np.random.uniform(-bound,bound,(N,d)))


def circle_draw(xc,yc,R,N):
    theta = np.linspace(0,2 * np.pi,N)
    X = np.cos(theta) * R + xc
    Y = np.sin(theta) * R + yc
    return X,Y


def random_walk_in_a_circle(x0 , y0 , xc , yc ,
                            R , N , step_size ,  param = 1 ,
                            polar = True , uniform_or_normal = 'n',
                            rand_seed = -1):
    """
    generate random walk in a circle

    Parameters
    ----------
    x0 : float
        x start of the random walk.
    y0 : float
        y start of the random walk.
    xc : float
        x center of the circle.
    yc : float
        y center of the circle.
    R : float
        radius of the circle.
    N : int
        number of points.
    step_size : float
        size of the step.
    param : int, optional
        control the random. The default is 1.
    polar : TYPE, optional
        DESCRIPTION. The default is True.
    uniform_or_normal : 'u' or 'n', optional
        uniform of normal walk. The default is 'n'.
    rand_seed : int, optional
        control the random. The default is -1.

    Returns
    -------
    X : np.array
        X-coordinates of the random walk.
    Y : np.array
        Y-coordinates of the random walk.
    Xcircle : np.array
        X-coordinates of the circle.
    Ycircle : np.array
        Y-coordinates of the circle.
        
    Exemple
    -------
    for un in ('u','n'):
    for pol in range(2):
        X,Y , Xcircle , Ycircle = random_walk_in_a_circle(10,10,0,0,50,10000,polar = pol,uniform_or_normal=un)

        plt.figure()
        plt.plot(Xcircle,Ycircle)
        plt.plot(X,Y)
        plt.axis('equal')
        plt.suptitle(un + str(pol))

    """

    X = [x0]
    Y = [y0]

    if rand_seed > -1:
        RAND = np.random.RandomState(rand_seed)
    else:
        RAND = np.random.RandomState(np.random.randint(10**6))

    Xcircle,Ycircle = circle_draw(xc,yc,R,500)

    for i in range(N-1):
        D = R+1
        iwhil = 0
        while D > R:
            iwhil += 1
            if iwhil > 500:
                log.warning('infinite loop in random_walk_in_a_circle ... %s')
            if polar:
                if uniform_or_normal == 'u':
                    dalpha = RAND.uniform(-param,param) * 2 * np.pi
                else:
                    dalpha = RAND.normal(0,param)       * 2 * np.pi
                drho = step_size
                dx = drho * np.cos(dalpha)
                dy = drho * np.sin(dalpha)
                #print dx,dy
            else:
                if uniform_or_normal == 'u':
                    dx = np.random.uniform(-param,param)
                    dy = np.random.uniform(-param,param)
                else:
                    dx = np.random.normal(0,param)
                    dy = np.random.normal(0,param)

            xtemp = X[-1] + dx
            ytemp = Y[-1] + dy
            D = np.sqrt((xtemp - xc)**2 + (ytemp - yc)**2)

        X.append(xtemp)
        Y.append(ytemp)

    X = np.array(X)
    Y = np.array(Y)

    return X,Y, Xcircle , Ycircle



def randn_bool(N,true_ratio = 0.5,RandGene = None):
    if RandGene is None:
        RandGene = np.random.RandomState()
    if type(RandGene) is int:
        RandGene = np.random.RandomState(RandGene)
    try:
        randlis = RandGene.uniform(size=N)
    except AttributeError:
        "ERR : AttributeError : RandGene  may be an int32/int64, but it's not an authentic int as required ..."
    boolout_lis = []
    for r in randlis:
        if r < true_ratio:
            boolout_lis.append(True)
        else:
            boolout_lis.append(False)

    return boolout_lis


def points_circle_border(Npts,r,r_sigma,az_type_normal=True,
                         main_dir=3.14159,dir_range=3.14159,seed=None):
    if not seed:
        seed = np.random.randint(10000)

    S = np.random.RandomState(seed)

    if not az_type_normal:
        Az = S.rand(Npts)  * 2 * np.pi
    else:
        Az = S.randn(Npts) * dir_range + main_dir


    R = np.array(Npts * [r]) - np.abs(S.randn(Npts) * r_sigma)

    X,Y = conv.polar2cartesian(R,Az,'rad')

    return X , Y


def estimated_autocorrelation(x):
    """
    http://stackoverflow.com/q/14297012/190597
    http://en.wikipedia.org/wiki/Autocorrelation#Estimation
    """
    n = len(x)
    variance = x.var()
    x = x-x.mean()
    r = np.correlate(x, x, mode = 'full')[-n:]
    assert np.allclose(r, np.array([(x[:n-k]*x[-(n-k):]).sum() for k in range(n)]))
    result = r/(variance*(np.arange(n, 0, -1)))
    return result

def savage_buford_formula(Vs,X,d):
    """
    X : distance à la faille , un iterable pour toutes le profil,
    un nombre pour la longeur max

    d : profondeur de la faille
    retourne X , et Vdeform(X)

    X et d doivent être dans la même unité, Vs pas forcément
    """

    if not utils.is_iterable(X):
        X = np.arange(-X,X,1)
    return X , ( Vs / np.pi ) * np.arctan2(X,d)




def R2_calc(y_obs,y_fit,with_r2_bis=False):
    #https://en.wikipedia.org/wiki/Coefficient_of_determination
    ybar = np.mean(y_obs)
    SStot = np.sum((y_obs - ybar)**2)
    SSreg = np.sum((y_fit - ybar)**2)
    SSres = np.sum((y_obs - y_fit)**2)

    r2 = 1. - ( SSres / SStot)
    r2bis = ( SSreg / SStot)

    if not with_r2_bis:
        return r2
    else:
        return r2 , r2bis

def R2_from_a_line_regress(Xobs,Yobs,a,b):
    #https://en.wikipedia.org/wiki/Coefficient_of_determination
    Xfit , Yfit = stats.linear_reg_getvalue(Xobs,a,b)
    r2 = R2_calc(Yobs,Yfit)
    return r2


def project_point_on_plan(N,M,A):
    """
    Project a point on a Plan

    Parameters
    ----------
    N : 3-Array (Vector)
        Vector describing the plan.
    M : 3-Array (Vector)
        Point of the plan.
    A : 3-Array (Vector)
        Point we want to project.

    Returns
    -------
    P : 3-Array (Vector)
        Projected point.
        
    Note
    ----
    It can be an ambiguity on the sign
    
    References
    ----------
    https://fr.wikipedia.org/wiki/Distance_d%27un_point_%C3%A0_un_plan
    https://www.mathematex.fr/viewtopic.php?t=923    
    """
    
    ##dist
    d = np.abs(N.dot(A-M)) / np.linalg.norm(N)
    
    ## Normalized N 
    Nnorm = N / np.linalg.norm(N) 
        
    P = Nnorm * d
    
    return P
    
 #  ______                _   _                _____                                         _ 
 # |  ____|              | | (_)              / ____|                                       | |
 # | |__ _   _ _ __   ___| |_ _  ___  _ __   | |  __ _ __ __ ___   _____ _   _  __ _ _ __ __| |
 # |  __| | | | '_ \ / __| __| |/ _ \| '_ \  | | |_ | '__/ _` \ \ / / _ \ | | |/ _` | '__/ _` |
 # | |  | |_| | | | | (__| |_| | (_) | | | | | |__| | | | (_| |\ V /  __/ |_| | (_| | | | (_| |
 # |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|  \_____|_|  \__,_| \_/ \___|\__, |\__,_|_|  \__,_|
 #                                                                        __/ |                
 #                                                                       |___/ 


def helmert_trans_legacy(Xa,params='itrf2008_2_etrf2000',
                         invert=True,workepoc=2009.):
    """
    This function is highly unstable and should be used only for legacy reason


    Parameters
    ----------
    Xa : TYPE
        DESCRIPTION.
    params : TYPE, optional
        DESCRIPTION. The default is 'itrf2008_2_etrf2000'.
    invert : TYPE, optional
        DESCRIPTION. The default is True.
    workepoc : TYPE, optional
        DESCRIPTION. The default is 2009..
    
    Returns
    -------
    Xb : TYPE
        DESCRIPTION.
        
    Notes
    -----
    NB 1 : http://etrs89.ensg.ign.fr/memo-V8.pdf
    NB 2 :
    Transformation inverse : * -1 pour les paramètres
    cf https://en.wikipedia.org/wiki/Helmert_transformation

    optimisé pour RGF93 => ITRF2008 (d'ou le invert = True & workepoc = 2009.)

    NB3 : Attention losque l'on compare avec
    le convertisseur EUREF avec une vitesse
    parce que elle aussi est modifiée dans la conversion ETRS => ITRS.
    Conclusion : le faire en 2 étapes: 
    ETRS => ITRS dans la même epoc
    ITRS epoc 1 => ITRS epoc 2
                                            
    NE MARCHE PAS PARCE BESOIN DU RATE(TAUX) DES PARAMS D'HELMERT !!!!!!
    (160923)
    
    if manual
    then params is a tuple
    params = (t1,t2,t3,dab,r1,r2,r3)

    """


    
    if invert:
        inver = -1.
    else:
        inver = 1.

    mas2rad = 0.0000000048481368
    mas2rad = 4.8481368111e-6 * 1e-3

    if params == 'itrf2008_2_etrf2000':

        t1rate = .1 * 10**-3
        t2rate = .1 * 10**-3
        t3rate = -1.8 * 10**-3
        dabrate = .08 * 10**-9
        r1rate =  .081 * mas2rad
        r2rate =  .490 * mas2rad
        r3rate = -.792 * mas2rad

        t1  =(52.1  * 10**-3   + t1rate  * ( workepoc - 2000.)) *inver
        t2  =(49.3  * 10**-3   + t2rate  * ( workepoc - 2000.)) *inver
        t3  =(-58.5 * 10**-3   + t3rate  * ( workepoc - 2000.)) *inver
        dab =( 1.34 * 10**-9   + dabrate * ( workepoc - 2000.)) *inver
        r1  =( 0.891 * mas2rad + r1rate  * ( workepoc - 2000.)) *inver
        r2  =( 5.39  * mas2rad + r2rate  * ( workepoc - 2000.)) *inver
        r3  =( -8.712* mas2rad + r3rate  * ( workepoc - 2000.)) *inver

    elif params == 'itrf2000_2_etrf2000':
        t1  =54.0  * 10**-3   *inver
        t2  =51.0  * 10**-3   *inver
        t3  =-48.0 * 10**-3   *inver
        dab = 0.0  * 10**-9   *inver
        r1  = 0.891 * mas2rad *inver
        r2  = 5.390 * mas2rad *inver
        r3  = -8.712* mas2rad *inver




    R = np.matrix([[dab,-r3,r2],
                   [r3,dab,-r1],
                   [-r2,r1,dab]])

    Xb = Xa + np.matrix([t1,t2,t3]) + np.dot(R,Xa)
    Xb = np.squeeze(np.array(Xb))

    return Xb



def mat_poids(Sinp,Ninp,fuvinp=1):
    """
    discontinued
    """
    # Sinp : liste des Sigmas sig = sqrt(var)
    # Ninp : liste de la taille de chaque blocs (obs)
    # fuvinp = 1 : facteur unitaire de variance

    if len(Sinp) != len(Ninp):
        raise Exception("S et N de taille differente")

    Ktemp = []

    for i in range(len(Sinp)):
        Ktemp.append(np.eye(Ninp[i]) * Sinp[i]**2)

    K = scipy.linalg.block_diag(*Ktemp)
    Q = (1/fuvinp) * K
    P = scipy.linalg.inv(Q)

    return K , Q , P



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: psakic
This sub-module of geodezyx.operational contains functions to download
gnss data and products from distant IGS servers. 
it can be imported directly with:
from geodezyx import operational
The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License
Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""

########## BEGIN IMPORT ##########
#### External modules
import datetime as dt
from ftplib import FTP, FTP_TLS
# import glob
import itertools
import multiprocessing as mp
import numpy as np
import os
# import pandas as pd 
import re
import shutil
# import urllib
# import ftplib

#### geodeZYX modules
from geodezyx import conv
from geodezyx import utils

import geodezyx.operational.download_utils as dlutils

#### Import star style
# from geodezyx import *                   # Import the GeodeZYX modules
# from geodezyx.externlib import *         # Import the external modules
# from geodezyx.megalib.megalib import *   # Import the legacy modules names


#### Import the logger
import logging
log = logging.getLogger(__name__)

##########  END IMPORT  ##########



#  _____               _            _         _____                      _                 _           
# |  __ \             | |          | |       |  __ \                    | |               | |          
# | |__) | __ ___   __| |_   _  ___| |_ ___  | |  | | _____      ___ __ | | ___   __ _  __| | ___ _ __ 
# |  ___/ '__/ _ \ / _` | | | |/ __| __/ __| | |  | |/ _ \ \ /\ / / '_ \| |/ _ \ / _` |/ _` |/ _ \ '__|
# | |   | | | (_) | (_| | |_| | (__| |_\__ \ | |__| | (_) \ V  V /| | | | | (_) | (_| | (_| |  __/ |   
# |_|   |_|  \___/ \__,_|\__,_|\___|\__|___/ |_____/ \___/ \_/\_/ |_| |_|_|\___/ \__,_|\__,_|\___|_|   
                                                                                                      
                                                                                                      

############################################################################
######## PRODUCTS DOWNLOADER
############################################################################

def multi_downloader_orbs_clks_2(archive_dir,startdate,enddate,
                                 AC_names = ("wum","cod"),
                                 prod_types = ("sp3","clk"),
                                 remove_patterns=("ULA",),
                                 archtype ='week',
                                 new_name_conv = True,
                                 parallel_download=4,
                                 archive_center='ign',
                                 mgex=True,
                                 repro=0,
                                 sorted_mode=False,
                                 return_also_uncompressed_files=True,
                                 ftp_download=False,
                                 dow_manu=False):
    """
    dow_manu = False, no dow manu, consider the converted dow from the time span, regular case
    dow_manu = None, no dow in the REGEX, the crawler will search only for the week
    dow_manu = 0 or 7: the dow in question    
    """
    
    if mgex:
        mgex_str = "mgex/"
    else:
        mgex_str = ""
        
    if repro:
        repro_str = "repro" + str(repro) + "/"
    else:
        repro_str = ""
        
    if not utils.is_iterable(remove_patterns):
        remove_patterns = [remove_patterns]
    
    secure_ftp = False
    
    log.info("data center used : %s",archive_center)
    log.info("mgex/repro : %s/%s",mgex,repro)
    
    if archive_center == "cddis":
        arch_center_main    = 'gdc.cddis.eosdis.nasa.gov'
        arch_center_basedir = '/pub/gps/products/' + mgex_str
        ftp_download = True
        secure_ftp = True
        parallel_download = 1
        log.info('cddis as data center, FTP and no parallel download forced')
        
    elif archive_center == "cddis_glonass":
        arch_center_main    = 'cddis.gsfc.nasa.gov'
        arch_center_basedir = '/pub/glonass/products/' + mgex_str
        
    elif archive_center == "ign":
        arch_center_main    = 'igs.ign.fr'
        arch_center_basedir = '/pub/igs/products/' + mgex_str  
        
    elif archive_center == "ign_iono":
        arch_center_main    = 'igs-rf.ign.fr'
        arch_center_basedir = '/pub/'  

    elif archive_center == "ensg":
        arch_center_main    = 'igs.ensg.ign.fr'
        arch_center_basedir = '/pub/igs/products/' + mgex_str
        
    elif archive_center == "whu":
        arch_center_main    = "igs.gnsswhu.cn"
        arch_center_basedir = "/pub/gps/products/" + mgex_str
        
    elif archive_center == "ign_rf":
        arch_center_main    = 'igs-rf.ign.fr'
        arch_center_basedir = '/pub/' + mgex_str  

    elif archive_center == "ensg_rf":
        arch_center_main    = 'igs-rf.ensg.ign.fr'
        arch_center_basedir = '/pub/' + mgex_str
        
    elif archive_center == "acc_xpr_mgex_cmb":
        ##### DO NOT WORK !!!!!
        ## and this archive is not mainteed anyway (2023-01)
        arch_center_main    = 'http://igsacc.s3-eu-central-1.amazonaws.com/products/mgex/final/2069/igm20694.sp3.Z'
        ftp_download = False
        log.info('ACC experimental mgex combi. as data center, HTTP download forced')


    Dates_list = conv.dt_range(startdate,enddate)

    wwww_dir_previous = None
    pool = mp.Pool(processes=parallel_download) 

    ## internal fct to create the FTP objects
    
    class MyFTP_TLS(FTP_TLS):
        """Explicit FTPS, with shared TLS session"""
        ### This new class is to avoid the error 
        ### ssl.SSLEOFError: EOF occurred in violation of protocol (_ssl.c:2396)
        ### source:
        ### https://stackoverflow.com/questions/14659154/ftps-with-python-ftplib-session-reuse-required
        def ntransfercmd(self, cmd, rest=None):
            conn, size = FTP.ntransfercmd(self, cmd, rest)
            if self._prot_p:
                conn = self.context.wrap_socket(conn,
                                                server_hostname=self.host,
                                                session=self.sock.session)  # this is the fix
            return conn, size
        
    def ftp_objt_create(secure_ftp_inp,chdir=""):
        
        # define the right constructor
        if secure_ftp_inp:
            ftp_constuctor = MyFTP_TLS
            #ftp=ftp_constuctor()
            #ftp.set_debuglevel(2)
            #ftp.connect(arch_center_main)
            #ftp.login('anonymous','')
            #ftp.prot_p()
        else:     
            ftp_constuctor = FTP
            #ftp = ftp_constuctor(arch_center_main)
            #ftp.login()
            
        ## create a list of FTP object for multiple downloads
        Ftp_obj_list_out = [ftp_constuctor(arch_center_main) for i in range(parallel_download)]
        if secure_ftp:
            [f.login('anonymous','') for f in Ftp_obj_list_out]    
            [f.prot_p() for f in Ftp_obj_list_out]    
        else:
            [f.login() for f in Ftp_obj_list_out]    
            
        # define the main obj for crawling
        ftp_main = Ftp_obj_list_out[0]
        
        # change the directory of the main ftp obj if we ask for it
        if chdir:
            log.info("Move to: %s",chdir)
            ftp_main.cwd(chdir)
        
        return ftp_main, Ftp_obj_list_out
    
    ###################################################################
    ########### Remote file search      

    Potential_localfiles_list_all = []
    
    ### check if the pattern of the wished products are in the listed daily files
    for ipatt_tup, patt_tup in enumerate(list(itertools.product(Dates_list,
                                                                AC_names,
                                                                prod_types))):
        dt_cur , ac_cur , prod_cur = patt_tup
        wwww , dow = conv.dt2gpstime(dt_cur)
        
        #### Manage the cases of manual DOW
        if type(dow_manu) is int:
            dow = dow_manu
        elif dow_manu is None:
            dow = ""
        elif dow_manu is False:
            pass        
        else:
            dow = str(dow_manu)
               
        log.info("*** Search prods. for %s-%s, AC/prod: %s/%s",wwww,dow,ac_cur,prod_cur)
        wwww_dir = os.path.join(arch_center_basedir,str(wwww),repro_str)
        
        n_ftp_ask = 500 ## Max interrogation of the FTP server to avoid 
                        ## potential errors
                        ## An new FTP instance is created if above it 
        
        if np.mod(ipatt_tup,n_ftp_ask) == 0:
            log.info("Create a new FTP instance")
            ftp, Ftp_obj_list = ftp_objt_create(secure_ftp)
            
        if wwww_dir_previous != wwww_dir or np.mod(ipatt_tup,n_ftp_ask) == 0:
            log.info("Move to: %s",wwww_dir)
            try:
                ftp.cwd(wwww_dir)
            except:
                log.warning("%s do not exists, skiping...",wwww_dir)
            Files_listed_in_FTP = ftp.nlst()
            wwww_dir_previous = wwww_dir
            if len(Files_listed_in_FTP) == 0:
                log.warning("no files found in directory %s",wwww_dir)
            
            
        Files_remote_date_list = []

        pattern_old_nam = ac_cur.lower()+".*"+str(wwww)+str(dow)+".*"+prod_cur.lower()+"\..*"
        Files = [f for f in Files_listed_in_FTP if re.search(pattern_old_nam,f)]
        
        pattern_new_nam = ""
        
        Files_new_nam = []
        if new_name_conv: ### search for new name convention
            
            if dow is None:
                log.error("dow == None and search for new name convention, Error ...")
                raise  Exception()
                
            ac_newnam   = ac_cur.upper()

            doy_newnam  = "".join(reversed(conv.dt2doy_year(conv.gpstime2dt(wwww,dow))))
            prod_newnam = prod_cur.upper()
            
            pattern_new_nam = utils.join_improved(".*",ac_newnam,doy_newnam,prod_newnam)
            pattern_new_nam = ".*" + pattern_new_nam + "\..*"

            Files_new_nam   = [f for f in Files_listed_in_FTP if re.search(pattern_new_nam,f)]
        
        
        log.info("Regex : %s %s",pattern_old_nam,pattern_new_nam)        
        Files = Files + Files_new_nam
        
        if len(Files) == 0:
            log.warning("no product found for" + " %s"*len(patt_tup),
                        *patt_tup)
        
        Files_remote_date_list = Files_remote_date_list + Files
            
        ### exclude some pattern
        for negpatt in remove_patterns:
            Files_remote_date_list = [e for e in Files_remote_date_list if not re.search(negpatt,e)]
            
            
        ###################################################################
        ########### Download
            

        archive_dir_specif = dlutils.effective_save_dir_orbit(archive_dir,
                                                              ac_cur,
                                                              dt_cur,
                                                              archtype)
        
        if len(Files_remote_date_list) > 0:
            utils.create_dir(archive_dir_specif)
        
        ### Generation of the Download fct inputs
        Files_remote_date_chunck = utils.chunkIt(Files_remote_date_list,
                                                 parallel_download)
        Downld_tuples_list = []
        Potential_localfiles_list = []

        if ftp_download: ### FTP Download
            for ftpobj , Chunk in zip(Ftp_obj_list,Files_remote_date_chunck):
                for filchunk in Chunk:
                    Potential_localfiles_list.append(os.path.join(archive_dir_specif,filchunk))
                    if parallel_download == 1:
                        Downld_tuples_list.append((ftpobj,filchunk,archive_dir_specif))
                    else:
                        Downld_tuples_list.append((arch_center_main,wwww_dir,
                                                   filchunk,archive_dir_specif))
        else: ### HTTP download
            Downld_tuples_list = itertools.product(["/".join(('ftp://' + arch_center_main,wwww_dir,f)) for f in Files_remote_date_list],[archive_dir_specif])
            [Potential_localfiles_list.append(os.path.join(archive_dir_specif,f)) for f in Files_remote_date_list]
        
        Potential_localfiles_list_all = Potential_localfiles_list_all +  Potential_localfiles_list 

        ### Actual Download
        if ftp_download and parallel_download == 1:
            for tup in Downld_tuples_list:
                dlutils.FTP_downloader(*tup)
        elif ftp_download and parallel_download > 1:
            _ = pool.map_async(dlutils.FTP_downloader_wo_objects,
                               Downld_tuples_list)
        elif not ftp_download and parallel_download == 1:
            for tup in Downld_tuples_list:
                dlutils.downloader_wrap(tup)
        elif not ftp_download and parallel_download > 1:
            _ = pool.map(dlutils.downloader_wrap,Downld_tuples_list)
    

    ###################################################################
    ########### Final Independent files existence check

    Localfiles_lis = []
    if not return_also_uncompressed_files:
        Pot_locfiles_list_use = Potential_localfiles_list_all
    else:
        Pot_locfiles_list_use = []
        for localfile in Potential_localfiles_list_all:
            Pot_compress_name_list = [localfile]
            Pot_compress_name_list.append(localfile.replace(".gz",""))
            Pot_compress_name_list.append(localfile.replace(".Z",""))
            Pot_compress_name_list = list(set(Pot_compress_name_list))
            
            Pot_locfiles_list_use = Pot_locfiles_list_use + Pot_compress_name_list
            
    for pot_localfile in Pot_locfiles_list_use:
        if os.path.isfile(pot_localfile):
            Localfiles_lis.append(pot_localfile)
    
    pool.close()
    return Localfiles_lis





def orbclk_long2short_name(longname_filepath_in,
                           rm_longname_file=False,
                           center_id_last_letter=None,
                           center_manual_short_name=None,
                           force=False,
                           dryrun=False,
                           output_dirname=None):
    """
    Rename a long naming new convention IGS product file to the short old
    convention
    Naming will be done automaticcaly based on the 3 first charaters of the
    long AC id
    e.g. CODE => cod, GRGS => grg, NOAA => noa ...

    Parameters
    ----------
    longname_filepath_in : str
        Full path of the long name product file

    rm_longname_file : bool
        Remove the original long name product file

    center_id_last_letter : str
        replace the last letter of the short AC id by another letter
        (see note below)
        
    center_manual_short_name : str
        replace completely the long name with this one
        overrides center_id_last_letter
        
    force : bool
        if False, skip if the file already exsists

    dryrun : bool
        if True, don't rename effectively, just output the new name
        
    output_dirname : str
        directory where the output shortname will be created
        if None, will be created in the same folder as the input longname

    Returns
    -------
    shortname_filepath : str
        Path of the  short old-named product file

    Note
    ----
    if you rename MGEX orbits, we advise to set
    center_id_last_letter="m"
    the AC code name will be changed to keep a MGEX convention
    (but any other caracter can be used too)

    e.g. for Bern's products, the long id is CODE

    if center_id_last_letter=None, it will become cod,
    if center_id_last_letter=m, it will become com

    """

    log.info("will rename " + longname_filepath_in)

    longname_basename = os.path.basename(longname_filepath_in)
    longname_dirname  = os.path.dirname(longname_filepath_in)
    
    if not output_dirname:
        output_dirname = longname_dirname

    center = longname_basename[:3]

    if center_manual_short_name:
        center = center_manual_short_name
    elif center_id_last_letter:
        center_as_list = list(center)
        center_as_list[-1] = center_id_last_letter
        center = "".join(center_as_list)

    yyyy   = int(longname_basename.split("_")[1][:4])
    doy    = int(longname_basename.split("_")[1][4:7])

    day_dt = conv.doy2dt(yyyy,doy)

    wwww , dow = conv.dt2gpstime(day_dt)

    shortname_prefix = center.lower() + str(wwww).zfill(4) + str(dow)

    ### Type handeling
    if   "SP3" in longname_basename:
        shortname = shortname_prefix + ".sp3"
    elif "CLK" in longname_basename:
        shortname = shortname_prefix + ".clk"
    elif "ERP" in longname_basename:
        shortname = shortname_prefix + ".erp"
    elif "BIA" in longname_basename:
        shortname = shortname_prefix + ".bia"
    elif "SNX" in longname_basename:
        shortname = shortname_prefix + ".snx"
    else:
        log.error("filetype not found for " + longname_basename)
    
    ### Compression handeling
    if longname_basename[-3:] == ".gz":
        shortname = shortname + ".gz"
    elif longname_basename[-2:] == ".Z":
        shortname = shortname + ".Z"
        
    shortname_filepath = os.path.join(output_dirname , shortname)
    
    if not force and os.path.isfile(shortname_filepath):
        log.info("skip " + longname_filepath_in)
        log.info(shortname_filepath + " already exists")        
        return shortname_filepath
    
    if not dryrun:
        log.info("renaming " + longname_filepath_in + " => " + shortname_filepath)
        shutil.copy2(longname_filepath_in , shortname_filepath)

    if rm_longname_file and not dryrun:
        log.info("remove " + longname_filepath_in)
        os.remove(longname_filepath_in)

    return shortname_filepath


 #  ______                _   _                _____                                         _ 
 # |  ____|              | | (_)              / ____|                                       | |
 # | |__ _   _ _ __   ___| |_ _  ___  _ __   | |  __ _ __ __ ___   _____ _   _  __ _ _ __ __| |
 # |  __| | | | '_ \ / __| __| |/ _ \| '_ \  | | |_ | '__/ _` \ \ / / _ \ | | |/ _` | '__/ _` |
 # | |  | |_| | | | | (__| |_| | (_) | | | | | |__| | | | (_| |\ V /  __/ |_| | (_| | | | (_| |
 # |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|  \_____|_|  \__,_| \_/ \___|\__, |\__,_|_|  \__,_|
 #                                                                        __/ |                
 #                                                                       |___/ 
 
 
def multi_downloader_orbs_clks(archive_dir,startdate,enddate,calc_center='igs',
                            sp3clk='sp3',archtype ='year/doy',parallel_download=4,
                            archive_center='ign',repro=0,sorted_mode=False,
                            force_weekly_file=False, return_also_uncompressed_files=True):

    """
    Download IGS products. Can manage MGEX products too
    (see archive_center argument)

    Parameters
    ----------
    archive_dir : str
        Parent archive directory where files will be stored

    startdate & enddate : datetime
        Start and End of the wished period

    calc_center : str or list of str
        calc_center can be a string or a list, describing the calc center
        e.g. 'igs','grg','cod','jpl' ...

    sp3clk : str
        Product type, can handle :

            'clk'

            'clk_30s'

            'sp3'
            
            'snx'
            
            'sum'

            'erp'

            'bia'

    archive_center : str
        server of download, "regular" IGS or MGEX, can handle :

            'cddis'

            'cddis_mgex'

            'cddis_mgex_longname'

            'ign'

            'ign_mgex'

            'ign_mgex_longname'
            
            'gfz_local'

    archtype: str
        string describing how the archive directory is structured, e.g :

            stat

            stat/year

            stat/year/doy

            year/doy

            year/stat

            week/dow/stat

            ... etc ...

    repro : int
        number of the IGS reprocessing (0 = routine processing)

    sorted_mode : bool
        if False:
            using the map multiprocess fct so the download order will
            be scrambled
        if True:
            using the apply multiprocess fct so the download order will be
            in the chrono. order
        The scrambled (False) is better, bc. it doesn't create zombies processes

    Returns
    -------
    localfiles_lis : list of str
        list of downloaded products paths
    """

    if type(calc_center) is str:
        calc_center =  [ ''.join([calc_center]) ] # POURQUOI CETTE LIGNE ?? (150717)
#        calc_center =  [calc_center]

    pool = mp.Pool(processes=parallel_download)
    urllist = []
    savedirlist = []

    if sp3clk == 'clk':
        typ = 'Clocks'
    elif sp3clk == 'clk_30s':
        typ = 'Clocks (30s)'
    elif sp3clk == 'sp3':
        typ = 'Orbits'
    elif sp3clk == 'erp':
        typ = 'ERP'
    elif sp3clk == 'snx':
        typ = 'SINEXs'
    elif sp3clk == 'sum':
        typ = 'SUM files'
    elif sp3clk == 'bia':
        typ = 'ISBs'
    else:
        typ = '????'

    log.info("generating the list of potential " + typ + " ...")
    
    for cc in calc_center:
            curdate = startdate
            while curdate <= enddate:
                if re.search("igs([0-9][0-9]|yy|YY)P",cc):
                    cc = "igs" + str(curdate.year)[2:] + "P"
                    log.info("INFO : IGS reference frame snx/ssc, correcting the year : " + cc)

                url = ''
                if archive_center == 'cddis':
                    url = orbclk_cddis_server(curdate,cc,repro=repro,sp3clk=sp3clk,
                                              force_weekly_file=force_weekly_file)
                elif archive_center == 'cddis_mgex':
                    url = orbclk_cddis_server(curdate,cc,repro=repro,sp3clk=sp3clk,
                                              mgex=True,force_weekly_file=force_weekly_file)
                elif archive_center == 'cddis_mgex_longname':
                    url = orbclk_cddis_server(curdate,cc,repro=repro,
                                              sp3clk=sp3clk,mgex=True,longname=True,
                                              force_weekly_file=force_weekly_file)
                elif archive_center == 'ign':
                    url = orbclk_ign_server(curdate,cc,repro=repro,sp3clk=sp3clk,
                                              mgex=False,force_weekly_file=force_weekly_file)
                elif archive_center == 'ign_mgex':
                    url = orbclk_ign_server(curdate,cc,repro=repro,sp3clk=sp3clk,
                                              mgex=True,force_weekly_file=force_weekly_file)
                elif archive_center == 'ign_mgex_longname':
                    url = orbclk_ign_server(curdate,cc,repro=repro,
                                              sp3clk=sp3clk,mgex=True,longname=True,
                                              force_weekly_file=force_weekly_file)
                elif archive_center == 'gfz_local':
                    url = orbclk_gfz_local_server(curdate,cc,repro=repro,
                                              sp3clk=sp3clk)

                else:
                    log.error('ERR : Wrong archive_center name !!! :' + archive_center)
                urllist.append(url)
                savedir = dlutils.effective_save_dir_orbit(archive_dir,
                                                           cc,
                                                           curdate,
                                                           archtype)
                savedirlist.append(savedir)
                curdate = curdate + dt.timedelta(days=1)

    savedirlist = [x for (y,x) in sorted(zip(urllist,savedirlist))]
    urllist = sorted(urllist)

    log.info(" ... done")
    log.info(str(len(urllist)) + " potential " + typ)

    if not sorted_mode:
        _ = pool.map(dlutils.downloader_wrap,list(zip(urllist,savedirlist)))
    else:
         results = [pool.apply_async(dlutils.downloader , args=(u,sd)) for u,sd in zip(urllist,savedirlist)]

    localfiles_lis = []
    
    if not return_also_uncompressed_files:
        for url , savedir in zip(urllist,savedirlist):
            localfile = os.path.join(savedir,os.path.basename(url))
            if os.path.isfile(localfile):
                localfiles_lis.append(localfile)
    else:
    
        for url , savedir in zip(urllist,savedirlist):

            localfile = os.path.join(savedir,os.path.basename(url))        
            
            Pot_compress_files_list = [localfile]
            Pot_compress_files_list.append(localfile.replace(".gz",""))
            Pot_compress_files_list.append(localfile.replace(".Z",""))
            Pot_compress_files_list = list(set(Pot_compress_files_list))
            
            for potential_exisiting_file in Pot_compress_files_list:
                if os.path.isfile(potential_exisiting_file):
                    localfiles_lis.append(potential_exisiting_file)
    
    pool.close()
    return localfiles_lis



def force_weekly_file_fct(force_weekly_file,sp3clk,day_in):
    if force_weekly_file == False:
        day = day_in

    elif type(force_weekly_file) is str or type(force_weekly_file) is int:
        log.info("INFO : The weekly file will be downloaded (DoW = %s)",force_weekly_file)
        log.info("       Check force_weekly_file option if you don't want it")
        day = force_weekly_file

    elif sp3clk in ("erp","sum") and force_weekly_file == True:
        log.info("The weekly file (DoW = 7) will be downloaded for " + sp3clk.upper())
        day = 7
        
    return day


def orbclk_cddis_server(date,center='igs', sp3clk = 'sp3', repro=0, mgex=False,
                        longname = False, force_weekly_file=False):
    """
    longname is experimental and only for MGEX yet !! (180426)
    """
    urlserver='ftp://cddis.gsfc.nasa.gov/pub/gps/products/'
    if mgex:
        urlserver = urlserver + 'mgex/'
    if repro == 0:
        rep_fldr = ''
    else:
        rep_fldr = 'repro' + str(repro)

    if repro != 0:
        center     = list(center)
        center[-1] = str(repro)
        center = ''.join(center)
    if repro == 3:
        longname = True
        
    if center in ("cod","cof","co2","cf2") and sp3clk == "sp3":
        log.info("CODE orbit extension changed to eph")
        sp3clk = "eph"

    # date definition
    week, day = conv.dt2gpstime(date)

    ## force_weekly_file handeling
    day = force_weekly_file_fct(force_weekly_file_fct,sp3clk,day)

    if not longname: # e.g. gbm19903.sp3.Z
        if not 'igu' in center:
            orbname = center + str(week).zfill(4)  + str(day) +'.'+ sp3clk +'.Z'
            url = os.path.join(urlserver, str(week).zfill(4) ,rep_fldr,orbname)
        else:
            igusuffix = center[-2:]
            center    = center[:-2]
            orbname = center + str(week).zfill(4)  + str(day)  + '_' + igusuffix +'.'+ sp3clk + '.Z'
            url = os.path.join(urlserver, str(week).zfill(4) ,rep_fldr,orbname)
    else: # e.g. COD0MGXFIN_20180580000_01D_05M_ORB.SP3.gz
        if len(center) == 3:
            center = center.upper() + "0"
        else:
            center = center.upper()

        datelong = date.strftime("%Y%j")

        if "SHA" in center:
            if sp3clk == "sp3":
                sp3clk_long = "_01D_15M_ORB.SP3"
            elif sp3clk == "clk":
                sp3clk_long = "_01D_05M_CLK.CLK"
            elif sp3clk == "erp":
                sp3clk_long = "_03D_12H_ERP.ERP"
            elif sp3clk == "bia":
                sp3clk_long = "_01D_01D_OSB.BIA"
            elif sp3clk == "snx":
                sp3clk_long = "_01D_000_SOL.SNX"

            orbname = center + "MGXRAP_" + datelong + "0000"  + sp3clk_long + ".gz"
        else:
            if sp3clk == "sp3":
                sp3clk_long = "_01D_15M_ORB.SP3"
            elif sp3clk == "clk":
                sp3clk_long = "_01D_30S_CLK.CLK"
            elif sp3clk == "erp":
                sp3clk_long = "_03D_12H_ERP.ERP"
            elif sp3clk == "bia":
                sp3clk_long = "_01D_01D_OSB.BIA"
            elif sp3clk == "snx":
                sp3clk_long = "_01D_000_SOL.SNX"

            orbname = center + "MGXFIN_" + datelong + "0000"  + sp3clk_long + ".gz"

        url = os.path.join(urlserver, str(week).zfill(4) ,rep_fldr,orbname)

    return url



def orbclk_ign_server(date,center='igs', sp3clk = 'sp3', repro=0, mgex=False,
                        longname = False,force_weekly_file=False):
    """
    longname is experimental and only for MGEX yet !! (180426)
    Must be merged with orbclk_cddis_server to make a big equivalent fct (180523)
    """
    urlserver='ftp://igs.ign.fr/pub/igs/products/'
    if mgex:
        urlserver = urlserver + 'mgex/'
    if repro == 0:
        rep_fldr = ''
    else:
        rep_fldr = 'repro' + str(repro)
    if repro != 0:
        center     = list(center)
        center[-1] = str(repro)
        center = ''.join(center)
    week, day = conv.dt2gpstime(date)

    ## force_weekly_file handeling
    day = force_weekly_file_fct(force_weekly_file,sp3clk,day)

    if not longname: # e.g. gbm19903.sp3.Z
        if not 'igu' in center:
            orbname = center + str(week).zfill(4)  + str(day) +'.'+ sp3clk +'.Z'
            url = os.path.join(urlserver, str(week).zfill(4) ,rep_fldr,orbname)
        else:
            igusuffix = center[-2:]
            center    = center[:-2]
            orbname = center + str(week).zfill(4)  + str(day)  + '_' + igusuffix +'.'+ sp3clk + '.Z'
            url = os.path.join(urlserver, str(week).zfill(4) ,rep_fldr,orbname)
    else: # e.g. COD0MGXFIN_20180580000_01D_05M_ORB.SP3.gz
        if len(center) == 3:
            center = center.upper() + "0"
        else:
            center = center.upper()

        datelong = date.strftime("%Y%j")

        if "SHA" in center:
            if sp3clk == "sp3":
                sp3clk_long = "_01D_15M_ORB.SP3"
            elif sp3clk == "clk":
                sp3clk_long = "_01D_05M_CLK.CLK"
            elif sp3clk == "erp":
                sp3clk_long = "_03D_12H_ERP.ERP"
            elif sp3clk == "bia":
                sp3clk_long = "_01D_01D_OSB.BIA"

            orbname = center + "MGXRAP_" + datelong + "0000"  + sp3clk_long + ".gz"
        else:
            if sp3clk == "sp3":
                sp3clk_long = "_01D_15M_ORB.SP3"
            elif sp3clk == "clk":
                sp3clk_long = "_01D_30S_CLK.CLK"
            elif sp3clk == "erp":
                sp3clk_long = "_03D_12H_ERP.ERP"
            elif sp3clk == "bia":
                sp3clk_long = "_01D_01D_OSB.BIA"

            orbname = center + "MGXFIN_" + datelong + "0000"  + sp3clk_long + ".gz"

        url = os.path.join(urlserver, str(week).zfill(4) ,rep_fldr,orbname)

    return url

def orbclk_igscb_server(date,center='gfz', sp3clk = 'sp3',repro=0):
    urlserver='ftp://igscb.jpl.nasa.gov/pub/product/'
    if repro == 0:
        rep_fldr = ''
    elif repro == 3:
        rep_fldr = 'repro' + str(repro)
    if repro != 0:
        center     = list(center)
        center[-1] = str(repro)
        center = ''.join(center)
    week, day = conv.dt2gpstime(date)
    if not 'igu' in center:
        orbname = center + str(week).zfill(4) + str(day) +'.' + sp3clk + '.Z'
        url = os.path.join(urlserver,str(week).zfill(4) ,rep_fldr,orbname)
    else:
        igusuffix = center[-2:]
        center    = center[:-2]
        orbname = center + str(week).zfill(4) + str(day)  + '_' + igusuffix +'.' + sp3clk + '.Z'
        url = os.path.join(urlserver,str(week).zfill(4) ,rep_fldr,orbname)
    return url

def orbclk_gfz_local_server(date,center='gfz', sp3clk='sp3',repro=0):
    if repro == 0:
        urlserver = '/dsk/igs_archive/IGS/SAVE_PROD_1d/'
    elif repro == 3:
        urlserver = '/dsk/repro3/ARCHIVE/IGS/SAVE_PROD_1d/'
    else:
        log.error("check the repro !!!")
        raise Exception

    week, day = conv.dt2gpstime(date)
    
    orbname = center + str(week).zfill(4) + str(day) +'.' + sp3clk + '.Z'
    url = os.path.join(urlserver,str(week).zfill(4),orbname)
    
    return url




